#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "list.h"

#define ELEMENTS 10

struct a_list {
	struct list_head list;
	int val;
	const char *str;
};

struct b_list {
	int val;
	struct list_head list;
	const char *str;
};

static char *rand_string(size_t size)
{
	const char charset[] = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";
	char *str = malloc(size + 1);
	if (size) {
		--size;
		for (size_t n = 0; n < size; n++) {
			int key = rand() % (int)(sizeof charset - 1);
			str[n] = charset[key];
		}
		str[size] = '\0';
	}
	return str;
}

static int rand_int(void)
{
	return rand();
}

static void append_a(struct a_list *ptr, const char *str, int val)
{
	struct a_list *tmp;
	tmp = (struct a_list *)malloc(sizeof(struct a_list));
	if (!tmp) {
		perror("malloc");
		exit(1);
	}
	tmp->str = str;
	tmp->val = val;
	list_add_tail(&(tmp->list), &(ptr->list));
}

static void append_b(struct b_list *ptr, const char *str, int val)
{
	struct b_list *tmp;
	tmp = (struct b_list *)malloc(sizeof(struct b_list));
	if (!tmp) {
		perror("malloc");
		exit(1);
	}
	tmp->str = str;
	tmp->val = val;
	list_add_tail(&(tmp->list), &(ptr->list));
}

int a_test(void)
{
	int i;
	struct a_list blist;
	struct a_list *iter;

	INIT_LIST_HEAD(&blist.list);

	/* add items to list */
	for (i = 0; i < ELEMENTS; i++) {
		append_a(&blist, rand_string(64), rand_int());
	}

	/* iterates list */
	list_for_each_entry(iter, &blist.list, list) {
		printf("%s %d\n", iter->str, iter->val);
	}

	/* remove all items in the list */
	while (!list_empty(&blist.list)) {
		iter = list_entry(blist.list.next, struct a_list, list);
		list_del(&iter->list);
		free(iter);
	}

	return 0;
}

int b_test(void)
{
	int i;
	struct b_list blist;
	struct b_list *iter;

	INIT_LIST_HEAD(&blist.list);

	/* add items to list */
	for (i = 0; i < ELEMENTS; i++) {
		append_b(&blist, rand_string(64), rand_int());
	}

	/* iterates list */
	list_for_each_entry(iter, &blist.list, list) {
		printf("%s %d\n", iter->str, iter->val);
	}

	/* remove all items in the list */
	while (!list_empty(&blist.list)) {
		iter = list_entry(blist.list.next, struct b_list, list);
		list_del(&iter->list);
		free(iter);
	}

	return 0;
}

int main(void)
{
	srand(time(0));

	puts("\n\t *** Test on struct A ***");
	a_test();
	puts("\n\t *** Test on struct B ***");
	b_test();
}

