#include <linux/slab.h>
#include <linux/time.h>
#include <asm/string.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/hrtimer.h>
#include <linux/ktime.h>

// https://www.kernel.org/doc/Documentation/timers/hrtimers.txt

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Alessandro Pellegrini <pellegrini@dis.uniroma1.it>");
MODULE_DESCRIPTION("High Resolution Timers Test");
MODULE_VERSION("1.0");

static struct hrtimer hr_timer;

unsigned long audit_counter = 0;
module_param(audit_counter, ulong, S_IRUSR | S_IRGRP | S_IROTH);

static enum hrtimer_restart enHRTimer = HRTIMER_NORESTART;
static s64 i64TimeInNsec = 100L;

enum hrtimer_restart my_hrtimer_callback(struct hrtimer *timer)
{
	audit_counter++;

	if(audit_counter % 1000000 == 0)
		printk(KERN_INFO "%s: Timer fired on CPU %d\n", KBUILD_MODNAME, smp_processor_id());

	hrtimer_forward(timer, hrtimer_cb_get_time(timer), ktime_set(0, i64TimeInNsec));   
	return enHRTimer;
}

static int __init timer_init(void)
{ 
	ktime_t kt;

	enHRTimer = HRTIMER_RESTART;

	//HRT init  
	kt = ktime_set(0, i64TimeInNsec);
	hrtimer_init(&hr_timer, CLOCK_REALTIME, HRTIMER_MODE_ABS);
	hrtimer_set_expires(&hr_timer, kt);
	hr_timer.function = &my_hrtimer_callback;
	hrtimer_start(&hr_timer, kt, HRTIMER_MODE_ABS);

	printk(KERN_INFO "%s: HR-Timer module loaded on CPU %d\n", KBUILD_MODNAME, smp_processor_id());
	return 0;
}

static void __exit timer_fini(void)
{
	//Reset
	hrtimer_cancel(&hr_timer);
	enHRTimer = HRTIMER_NORESTART;
	printk(KERN_INFO "%s: HR-Timer module unloaded on CPU %d\n", KBUILD_MODNAME, smp_processor_id());
}

module_init(timer_init)
module_exit(timer_fini)
