#include <linux/module.h>

static atomic_t synch_leave;
static atomic_t synch_enter;

static void synchronize_all_slaves(void *info) {
	(void)info;
	preempt_disable();
	printk(KBUILD_MODNAME ": cpu %d entering synchronize_all_slaves\n", smp_processor_id());
	atomic_dec(&synch_enter);
	while(atomic_read(&synch_leave) > 0);
	printk(KBUILD_MODNAME ": cpu %d leaving synchronize_all_slaves\n", smp_processor_id());
	preempt_enable();
}

void synchronize_all(void) {
	preempt_disable();
	printk(KBUILD_MODNAME ": cpu %d asking from unpreemptive synchronization\n", smp_processor_id());
	atomic_set(&synch_enter, num_online_cpus() - 1);
	atomic_set(&synch_leave, 1);
	smp_call_function_many(cpu_online_mask, synchronize_all_slaves, NULL , false);
	while(atomic_read(&synch_enter) > 0);
	printk(KBUILD_MODNAME ": cpu %d all kernel threads synchronized\n", smp_processor_id());
}


void unsynchronize_all(void) {
	printk(KBUILD_MODNAME ": cpu %d freeing other kernel threads\n", smp_processor_id());
	atomic_set(&synch_leave, 0);
	preempt_enable();
}

static int __init synch_init(void)
{
	printk(KBUILD_MODNAME ": Initializing synchronization...");
	synchronize_all();
	printk(KBUILD_MODNAME ": Running in isolation");
	unsynchronize_all();

	return 0;
}

static void __exit synch_cleanup(void)
{
}


MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Synchronize all cores upon kernel module");

module_init(synch_init);
module_exit(synch_cleanup);

