#include <linux/sched.h>
#include <linux/module.h>
#include <linux/syscalls.h>
#include <linux/version.h>

#include "lend.h"


#if (LINUX_VERSION_CODE < KERNEL_VERSION(4,17,0))
unsigned char *get_syscall_table(void)
{
	unsigned long *syscall_table;
	unsigned long int i;

	for (i = (unsigned long int)close; i < ULONG_MAX; i += sizeof(void *)) {
		syscall_table = (unsigned long *)i;

		if (syscall_table[__NR_close] == (unsigned long)sys_close)
			return syscall_table;
	}
	return NULL;
}
#else

#define SEARCH_RANGE 512
unsigned char *memmem(const void *haystack, size_t haystacklen, const void *needle, size_t needlelen)
{
    char *p;

    for (p = (char *)haystack; p <= ((char *)haystack - needlelen + haystacklen); p++)
        if (memcmp(p, needle, needlelen) == 0)
            return (void *)p;

    return NULL;
}

static unsigned char *find_do_syscall_64(void) {
	unsigned char *do_syscall_64 = NULL;
	unsigned char *entry_SYSCALL_64 = NULL;

	unsigned char pattern[] = {0x48, 0x89, 0xc7, 0x48, 0x89, 0xe6, 0xe8};

	entry_SYSCALL_64 = (unsigned char *)native_read_msr(MSR_LSTAR);

	do_syscall_64 = memmem(entry_SYSCALL_64, SEARCH_RANGE, pattern, sizeof(pattern)) + sizeof(pattern);
	if (!do_syscall_64) {
		return NULL;
	}

	do_syscall_64 = do_syscall_64 + 4 + (int64_t)*((int32_t *)do_syscall_64);

	return do_syscall_64;
}

#ifndef CONFIG_RETPOLINE
static unsigned char *get_table(unsigned char *base)
{
	unsigned char *sys_call_table = NULL;

	unsigned char pattern[] = {0x48, 0x89, 0xef, 0xff, 0x14, 0xc5};

	sys_call_table = memmem(base, SEARCH_RANGE, pattern, sizeof(pattern)) + sizeof(pattern);
	if (!sys_call_table) {
		return NULL;
	}

	return sys_call_table;
}
#else

unsigned char *get_table(unsigned char *base)
{
	unsigned char *syscall_table = NULL;
	unsigned int size;
	unsigned char *p = base;

	while(p < base + SEARCH_RANGE) {
		size = length_disasm(p, MODE_X64);
		if(size == 8 && (p[0] & 0x48) == 0x48 // Catches both 0x48 and 0x4c
			     && p[1] == 0x8b 
			     && (p[2] & 0xc0) == 0x00 // mod = 00
			     && (p[2] & 0x04) == 0x04) { // r/m == 100
			break;
		}

		p += size;
	}

	if(p < base + SEARCH_RANGE) {
		syscall_table = (unsigned char *)(int64_t)*((int32_t *)(p + 4));
	}
	
	if (!syscall_table) {
		return NULL;
	}
	return syscall_table;
	
}
#endif /* CONFIG_RETPOLINE */

unsigned char *get_syscall_table(void)
{
	unsigned char *do_syscall_64;
	unsigned char *table = NULL;

	do_syscall_64 = find_do_syscall_64();

	if(do_syscall_64)
		table = get_table(do_syscall_64);

	return table;
}
#endif /* (LINUX_VERSION_CODE < KERNEL_VERSION(4,17,0)) */

